# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
# flake8: noqa
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

#################
# MRT CONSTANTS #
#################

# Byte sizes
MRT_HEADER_BYTES = 12     # Defined in RFC6396
MRT_DATA_BYTES   = 65535  # Defined in RFC8654
MRT_DATA_BYTES  *= 16     # RIB entries might be larger

# MRT types
MRT_NULL          = 0   # Deprecated in RFC6396
MRT_START         = 1   # Deprecated in RFC6396
MRT_DIE           = 2   # Deprecated in RFC6396
MRT_I_AM_DEAD     = 3   # Deprecated in RFC6396
MRT_PEER_DOWN     = 4   # Deprecated in RFC6396
MRT_BGP           = 5   # Deprecated in RFC6396
MRT_RIP           = 6   # Deprecated in RFC6396
MRT_IDRP          = 7   # Deprecated in RFC6396
MRT_RIPNG         = 8   # Deprecated in RFC6396
MRT_BGP4PLUS      = 9   # Deprecated in RFC6396
MRT_BGP4PLUS_01   = 10  # Deprecated in RFC6396
MRT_OSPFV2        = 11  # Defined in RFC6396
MRT_TABLE_DUMP    = 12  # Defined in RFC6396
MRT_TABLE_DUMP_V2 = 13  # Defined in RFC6396
MRT_BGP4MP        = 16  # Defined in RFC6396
MRT_BGP4MP_ET     = 17  # Defined in RFC6396
MRT_ISIS          = 32  # Defined in RFC6396
MRT_ISIS_ET       = 33  # Defined in RFC6396
MRT_OSPFV3        = 48  # Defined in RFC6396
MRT_OSPFV3_ET     = 49  # Defined in RFC6396

# BGP subtypes (obsolete)
MRT_BGP_NULL         = 0  # Deprecated in RFC6396
MRT_BGP_UPDATE       = 1  # Deprecated in RFC6396
MRT_BGP_PREF_UPDATE  = 2  # Deprecated in RFC6396
MRT_BGP_STATE_CHANGE = 3  # Deprecated in RFC6396
MRT_BGP_SYNC         = 4  # Deprecated in RFC6396
MRT_BGP_OPEN         = 5  # Deprecated in RFC6396
MRT_BGP_NOTIFY       = 6  # Deprecated in RFC6396
MRT_BGP_KEEPALIVE    = 7  # Deprecated in RFC6396

# BGP4MP/BGP4MP_ET subtypes
MRT_BGP4MP_STATE_CHANGE              = 0   # Defined in RFC6396
MRT_BGP4MP_MESSAGE                   = 1   # Defined in RFC6396
MRT_BGP4MP_ENTRY                     = 2   # Deprecated in RFC6396
MRT_BGP4MP_SNAPSHOT                  = 3   # Deprecated in RFC6396
MRT_BGP4MP_MESSAGE_AS4               = 4   # Defined in RFC6396
MRT_BGP4MP_STATE_CHANGE_AS4          = 5   # Defined in RFC6396
MRT_BGP4MP_MESSAGE_LOCAL             = 6   # Defined in RFC6396
MRT_BGP4MP_MESSAGE_AS4_LOCAL         = 7   # Defined in RFC6396
MRT_BGP4MP_MESSAGE_ADDPATH           = 8   # Defined in RFC8050
MRT_BGP4MP_MESSAGE_AS4_ADDPATH       = 9   # Defined in RFC8050
MRT_BGP4MP_MESSAGE_LOCAL_ADDPATH     = 10  # Defined in RFC8050
MRT_BGP4MP_MESSAGE_AS4_LOCAL_ADDPATH = 11  # Defined in RFC8050

# TABLE_DUMP_V2 subtypes
MRT_TABLE_DUMP_V2_PEER_INDEX_TABLE           = 1   # Defined in RFC6396
MRT_TABLE_DUMP_V2_RIB_IPV4_UNICAST           = 2   # Defined in RFC6396
MRT_TABLE_DUMP_V2_RIB_IPV4_MULTICAST         = 3   # Defined in RFC6396
MRT_TABLE_DUMP_V2_RIB_IPV6_UNICAST           = 4   # Defined in RFC6396
MRT_TABLE_DUMP_V2_RIB_IPV6_MULTICAST         = 5   # Defined in RFC6396
MRT_TABLE_DUMP_V2_RIB_GENERIC                = 6   # Defined in RFC6396
MRT_TABLE_DUMP_V2_GEO_PEER_TABLE             = 7   # Defined in RFC6397
MRT_TABLE_DUMP_V2_RIB_IPV4_UNICAST_ADDPATH   = 8   # Defined in RFC8050
MRT_TABLE_DUMP_V2_RIB_IPV4_MULTICAST_ADDPATH = 9   # Defined in RFC8050
MRT_TABLE_DUMP_V2_RIB_IPV6_UNICAST_ADDPATH   = 10  # Defined in RFC8050
MRT_TABLE_DUMP_V2_RIB_IPV6_MULTICAST_ADDPATH = 11  # Defined in RFC8050
MRT_TABLE_DUMP_V2_RIB_GENERIC_ADDPATH        = 12  # Defined in RFC8050

#####################
# MRT CONSTANT SETS #
#####################

# Valid MRT types
MRT_VALID = {
    MRT_NULL,
    MRT_START,
    MRT_DIE,
    MRT_I_AM_DEAD,
    MRT_PEER_DOWN,
    MRT_BGP,
    MRT_RIP,
    MRT_IDRP,
    MRT_RIPNG,
    MRT_BGP4PLUS,
    MRT_BGP4PLUS_01,
    MRT_OSPFV2,
    MRT_TABLE_DUMP,
    MRT_TABLE_DUMP_V2,
    MRT_BGP4MP,
    MRT_BGP4MP_ET,
    MRT_ISIS,
    MRT_ISIS_ET,
    MRT_OSPFV3,
    MRT_OSPFV3_ET,
}

# BGP4MP/BGP4MP_ET subtypes (2-byte ASN)
MRT_BGP4MP_ENTRY_ANY = {
    MRT_BGP4MP_STATE_CHANGE,
    MRT_BGP4MP_MESSAGE,
    MRT_BGP4MP_MESSAGE_LOCAL,
    MRT_BGP4MP_MESSAGE_ADDPATH,
    MRT_BGP4MP_MESSAGE_LOCAL_ADDPATH,
}

# BGP4MP/BGP4MP_ET subtypes (4-byte ASN)
MRT_BGP4MP_ENTRY_AS4_ANY = {
    MRT_BGP4MP_STATE_CHANGE_AS4,
    MRT_BGP4MP_MESSAGE_AS4,
    MRT_BGP4MP_MESSAGE_AS4_LOCAL,
    MRT_BGP4MP_MESSAGE_AS4_ADDPATH,
    MRT_BGP4MP_MESSAGE_AS4_LOCAL_ADDPATH,
}

# BGP4MP/BGP4MP_ET subtypes (ADD-PATH)
MRT_BGP4MP_ENTRY_ADDPATH_ANY = {
    MRT_BGP4MP_MESSAGE_ADDPATH,
    MRT_BGP4MP_MESSAGE_AS4_ADDPATH,
    MRT_BGP4MP_MESSAGE_LOCAL_ADDPATH,
    MRT_BGP4MP_MESSAGE_AS4_LOCAL_ADDPATH,
}

# TABLE_DUMP_V2 subtypes (ADD-PATH)
MRT_TABLE_DUMP_V2_RIB_ADDPATH_ANY = {
    MRT_TABLE_DUMP_V2_RIB_IPV4_UNICAST_ADDPATH,
    MRT_TABLE_DUMP_V2_RIB_IPV4_MULTICAST_ADDPATH,
    MRT_TABLE_DUMP_V2_RIB_IPV6_UNICAST_ADDPATH,
    MRT_TABLE_DUMP_V2_RIB_IPV6_MULTICAST_ADDPATH,
    MRT_TABLE_DUMP_V2_RIB_GENERIC_ADDPATH,
}

# TABLE_DUMP_V2 subtypes (IPv4)
MRT_TABLE_DUMP_V2_RIB_IPV4_ANY = {
    MRT_TABLE_DUMP_V2_RIB_IPV4_UNICAST,
    MRT_TABLE_DUMP_V2_RIB_IPV4_MULTICAST,
    MRT_TABLE_DUMP_V2_RIB_IPV4_UNICAST_ADDPATH,
    MRT_TABLE_DUMP_V2_RIB_IPV4_MULTICAST_ADDPATH,
}

# TABLE_DUMP_V2 subtypes (IPv6)
MRT_TABLE_DUMP_V2_RIB_IPV6_ANY = {
    MRT_TABLE_DUMP_V2_RIB_IPV6_UNICAST,
    MRT_TABLE_DUMP_V2_RIB_IPV6_MULTICAST,
    MRT_TABLE_DUMP_V2_RIB_IPV6_UNICAST_ADDPATH,
    MRT_TABLE_DUMP_V2_RIB_IPV6_MULTICAST_ADDPATH,
}

# TABLE_DUMP_V2 subtypes (generic)
MRT_TABLE_DUMP_V2_RIB_GENERIC_ANY = {
    MRT_TABLE_DUMP_V2_RIB_GENERIC,
    MRT_TABLE_DUMP_V2_RIB_GENERIC_ADDPATH,
}

# TABLE_DUMP_V2 subtypes (AFI/SAFI-specific)
MRT_TABLE_DUMP_V2_RIB_SPECIFIC_ANY = (
    MRT_TABLE_DUMP_V2_RIB_IPV4_ANY |
    MRT_TABLE_DUMP_V2_RIB_IPV6_ANY
)



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# System imports
import base64
import json

# Local imports
from ..model.attr import FTL_ATTR_BGP_STATS_MRT_ERRORS
from ..model.attr import FTL_ATTR_BGP_STATS_LGL_ERRORS
from ..model.attr import FTL_ATTR_BGP_STATS_PARSER_ERRORS
from ..model.attr import FTL_ATTR_BGP_ERROR_RECORD
from ..model.attr import FTL_ATTR_BGP_ERROR_SOURCE
from ..model.attr import FTL_ATTR_BGP_ERROR_SCOPE
from ..model.attr import FTL_ATTR_BGP_ERROR_REASON
from ..model.attr import FTL_ATTR_BGP_ERROR_MESSAGE
from ..model.attr import FTL_ATTR_BGP_ERROR_DATA
from ..model.attr import FTL_ATTR_BGP_ERROR_TRACE
from ..model.attr import FTL_ATTR_BGP_ERROR_SOURCE_HUMAN
from ..model.attr import FTL_ATTR_BGP_ERROR_SCOPE_HUMAN
from ..model.attr import FTL_ATTR_BGP_ERROR_REASON_HUMAN
from ..model.attr import FTL_ATTR_BGP_ERROR_DATA_HUMAN
from ..model.const import FTL_ATTR_BGP_ERROR_SOURCE_MRT
from ..model.const import FTL_ATTR_BGP_ERROR_SOURCE_LGL
from ..model.const import FTL_ATTR_BGP_ERROR_SOURCE_PARSER
from ..model.const import FTL_ATTR_BGP_ERROR_SOURCE_TO_STR
from ..model.const import FTL_ATTR_BGP_ERROR_SCOPE_TO_STR
from ..model.const import FTL_ATTR_BGP_ERROR_REASON_TO_STR
from ..model.record import FTL_RECORD_BGP_STATS
from ..model.record import FTL_RECORD_BGP_ERROR

# Cache keys
CACHE_TS = 0


def init_caches():
    """ Initialize data caches.
    """
    # Return data caches
    # NOTE: We currently support timestamp caches only
    return (dict(), )


def handle_bgp_error(raise_on_errors, error_init, error_emit, error_type, stats_emit, stats_record, error):
    """ Yield BGP error record or re-raise given exception.
    """
    # Extract error source
    error_source = error.source
    if FTL_ATTR_BGP_ERROR_SOURCE_HUMAN:
        error_source = FTL_ATTR_BGP_ERROR_SOURCE_TO_STR.get(error_source, str(error_source))

    # Extract error scope
    error_scope = error.scope
    if FTL_ATTR_BGP_ERROR_SCOPE_HUMAN:
        error_scope = FTL_ATTR_BGP_ERROR_SCOPE_TO_STR.get(error_scope, str(error_scope))

    # Extract error reason
    error_reason = error.reason
    if FTL_ATTR_BGP_ERROR_REASON_HUMAN:
        error_reason = FTL_ATTR_BGP_ERROR_REASON_TO_STR.get(error_reason, str(error_reason))

    # Update stats record
    if FTL_RECORD_BGP_STATS:

        # Identify error source
        source_errors = None
        if error.source == FTL_ATTR_BGP_ERROR_SOURCE_MRT:
            source_errors = FTL_ATTR_BGP_STATS_MRT_ERRORS
        elif error.source == FTL_ATTR_BGP_ERROR_SOURCE_LGL:
            source_errors = FTL_ATTR_BGP_STATS_LGL_ERRORS
        elif error.source == FTL_ATTR_BGP_ERROR_SOURCE_PARSER:
            source_errors = FTL_ATTR_BGP_STATS_PARSER_ERRORS

        # Add error scope/record/reason
        if source_errors:
            error_str = '|'.join((str(err) for err in (error_scope, error_reason, error_type)))
            stats_record[source_errors][error_str] = stats_record[source_errors].get(error_str, 0) + 1

        # Yield final stats record (if exception will be raised)
        if raise_on_errors:
            yield stats_emit(stats_record)

    # Yield error record if requested
    if FTL_RECORD_BGP_ERROR:

        # Initialize error record
        error_record = list(error_init)

        # Add error source to record
        if FTL_ATTR_BGP_ERROR_SOURCE >= 0:
            error_record[FTL_ATTR_BGP_ERROR_SOURCE] = error_source

        # Add error scope to record
        if FTL_ATTR_BGP_ERROR_SCOPE >= 0:
            error_record[FTL_ATTR_BGP_ERROR_SCOPE] = error_scope

        # Add error record type to record
        if FTL_ATTR_BGP_ERROR_RECORD >= 0:
            error_record[FTL_ATTR_BGP_ERROR_RECORD] = error_type

        # Add error reason to record
        if FTL_ATTR_BGP_ERROR_REASON >= 0:
            error_record[FTL_ATTR_BGP_ERROR_REASON] = error_reason

        # Add error message to record
        if FTL_ATTR_BGP_ERROR_MESSAGE >= 0:
            error_record[FTL_ATTR_BGP_ERROR_MESSAGE] = error.message

        # Add error data to record
        if FTL_ATTR_BGP_ERROR_DATA >= 0:
            data = error.data
            if data is not None:
                if isinstance(data, (bytes, bytearray, memoryview)) is True:
                    if FTL_ATTR_BGP_ERROR_DATA_HUMAN:
                        data = ' '.join('{:02x}'.format(byte) for byte in data)
                    else:
                        data = base64.b64encode(data).decode('ascii')
                elif FTL_ATTR_BGP_ERROR_DATA_HUMAN:
                    data = json.dumps(data)
            error_record[FTL_ATTR_BGP_ERROR_DATA] = data

        # Add error trace to record
        if FTL_ATTR_BGP_ERROR_TRACE >= 0:
            error_record[FTL_ATTR_BGP_ERROR_TRACE] = error.trace

        # Yield final error record
        yield error_emit(error_record)

    # Re-raise exception if requested
    if raise_on_errors:

        # Clone and raise error
        raise type(error)(record=error_type, error=error)



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# System imports
from collections import namedtuple

# Local imports
from .attr import FtlAttrsBgp
from .util import FTL_ATTR_SHIFT_HUMAN


def filter_stats_spec(no_support):
    """ Remove unsupported stats fields from attributes spec.
    """
    # Update supported stats attributes
    stats, stats_fields, stats_human = FtlAttrsBgp.stats, dict(), None
    update_fields, unsupported_fields = list(), 0
    for attr_name, attr_value in stats._asdict().items():
        if attr_value != stats.human:
            if attr_value in {stats.ALL, stats.NONE, stats.DEFAULT}:
                update_fields.append(attr_name)
                stats_fields[attr_name] = attr_value
            elif attr_name in stats.internal or any(attr_name.startswith(skip) for skip in no_support) is False:
                stats_fields[attr_name] = attr_value
            else:
                unsupported_fields |= attr_value
        else:
            stats_fields[attr_name] = attr_value
            stats_human = attr_name

    # Update supported stats attributes (human-readable)
    stats_fields_human = {field: value << FTL_ATTR_SHIFT_HUMAN for field, value in stats_fields.items()
                          if field not in stats.internal or field in update_fields}
    for field in update_fields:
        stats_fields[field] &= ~unsupported_fields
        stats_fields_human[field] = stats_fields[field] << FTL_ATTR_SHIFT_HUMAN
    stats_fields[stats_human] = namedtuple(stats.human.__class__.__name__,
                                           stats_fields_human.keys())(*stats_fields_human.values())

    # Clone and return filtered attributes spec
    attrs = FtlAttrsBgp._asdict()
    attrs[stats.name] = namedtuple(stats.__class__.__name__, stats_fields.keys())(*stats_fields.values())
    return namedtuple(FtlAttrsBgp.__class__.__name__, attrs.keys())(*attrs.values())


def init_stats_record(attributes, attrindexes, spec_attributes, stats_record):
    """ Initialize default values for stats record.
    """
    # Initialize parser stats values
    try:
        if spec_attributes & (attributes.parser_lifetime | attributes.human.parser_lifetime):
            stats_record[attrindexes.parser_lifetime] = 0.0
        if spec_attributes & (attributes.parser_runtime | attributes.human.parser_runtime):
            stats_record[attrindexes.parser_runtime] = 0.0
        if spec_attributes & (attributes.parser_errors | attributes.human.parser_errors):
            stats_record[attrindexes.parser_errors] = dict()
    except AttributeError:
        pass

    # Initialize looking-glass stats values
    try:
        if spec_attributes & (attributes.lgl_runtime | attributes.human.lgl_runtime):
            stats_record[attrindexes.lgl_runtime] = 0.0
        if spec_attributes & (attributes.lgl_entries | attributes.human.lgl_entries):
            stats_record[attrindexes.lgl_entries] = 0
        if spec_attributes & (attributes.lgl_errors | attributes.human.lgl_errors):
            stats_record[attrindexes.lgl_errors] = dict()
    except AttributeError:
        pass

    # Initialize MRT stats values
    try:
        if spec_attributes & (attributes.mrt_runtime | attributes.human.mrt_runtime):
            stats_record[attrindexes.mrt_runtime] = 0.0
        if spec_attributes & (attributes.mrt_entries | attributes.human.mrt_entries):
            stats_record[attrindexes.mrt_entries] = 0
        if spec_attributes & (attributes.mrt_errors | attributes.human.mrt_errors):
            stats_record[attrindexes.mrt_errors] = dict()
        if spec_attributes & (attributes.mrt_fixes | attributes.human.mrt_fixes):
            stats_record[attrindexes.mrt_fixes] = dict()
        if spec_attributes & (attributes.mrt_bgp_entry_types | attributes.human.mrt_bgp_entry_types):
            stats_record[attrindexes.mrt_bgp_entry_types] = dict()
        if spec_attributes & (attributes.mrt_bgp_message_types | attributes.human.mrt_bgp_message_types):
            stats_record[attrindexes.mrt_bgp_message_types] = dict()
        if spec_attributes & (attributes.mrt_bgp_attribute_types | attributes.human.mrt_bgp_attribute_types):
            stats_record[attrindexes.mrt_bgp_attribute_types] = dict()
        if spec_attributes & (attributes.mrt_bgp_capability_types | attributes.human.mrt_bgp_capability_types):
            stats_record[attrindexes.mrt_bgp_capability_types] = dict()
    except AttributeError:
        pass

    # Initialize BGP stats values
    try:
        if spec_attributes & (attributes.bgp_routes_rib_ipv4 | attributes.human.bgp_routes_rib_ipv4):
            stats_record[attrindexes.bgp_routes_rib_ipv4] = 0
        if spec_attributes & (attributes.bgp_routes_rib_ipv6 | attributes.human.bgp_routes_rib_ipv6):
            stats_record[attrindexes.bgp_routes_rib_ipv6] = 0
        if spec_attributes & (attributes.bgp_routes_announce_ipv6 | attributes.human.bgp_routes_announce_ipv6):
            stats_record[attrindexes.bgp_routes_announce_ipv6] = 0
        if spec_attributes & (attributes.bgp_routes_announce_ipv4 | attributes.human.bgp_routes_announce_ipv4):
            stats_record[attrindexes.bgp_routes_announce_ipv4] = 0
        if spec_attributes & (attributes.bgp_routes_withdraw_ipv4 | attributes.human.bgp_routes_withdraw_ipv4):
            stats_record[attrindexes.bgp_routes_withdraw_ipv4] = 0
        if spec_attributes & (attributes.bgp_routes_withdraw_ipv6 | attributes.human.bgp_routes_withdraw_ipv6):
            stats_record[attrindexes.bgp_routes_withdraw_ipv6] = 0
    except AttributeError:
        pass
